# Copyright © 2015 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.


import os
from score.init import ConfiguredModule, ConfigurationError
from .process import (
    Overlord, Zergling, NoSuchZergling, AlreadyPaused,
    AlreadyRunning, AlreadyReloading, NotRunning)


defaults = {
    'rootdir': None,
}


def init(confdict):
    """
    Initializes this module acoording to :ref:`our module initialization
    guidelines <module_initialization>` with the following configuration keys:

    :confkey:`rootdir` :faint:`[default=None]`
        The folder containing all uwsgi instances' files.

    """
    conf = defaults.copy()
    conf.update(confdict)
    if not conf['rootdir']:
        raise ConfigurationError(__package__,
                                 'No root folder provided')
    os.makedirs(conf['rootdir'], exist_ok=True)
    return ConfiguredUwsgiModule(conf['rootdir'])


class ConfiguredUwsgiModule(ConfiguredModule):
    """
    This module's :class:`configuration object
    <score.init.ConfiguredModule>`.
    """

    def __init__(self, rootdir):
        super().__init__(__package__)
        self.rootdir = rootdir
        conf = self

        class ConfiguredOverlord(Overlord):

            @classmethod
            def instances(cls):
                for folder in os.listdir(conf.rootdir):
                    if os.path.isdir(os.path.join(conf.rootdir, folder)):
                        overlord = cls(folder)
                        if overlord.is_running():
                            yield overlord

            def __init__(self, *args, **kwargs):
                self.conf = conf
                super().__init__(*args, **kwargs)

        class ConfiguredZergling(Zergling):

            def __init__(self, *args, **kwargs):
                self.conf = conf
                super().__init__(*args, **kwargs)

        self.Overlord = ConfiguredOverlord
        self.Zergling = ConfiguredZergling


__all__ = [
    'init', 'ConfiguredUwsgiModule', 'Overlord', 'Zergling', 'NoSuchZergling',
    'AlreadyPaused', 'AlreadyRunning', 'AlreadyReloading', 'NotRunning']



# Copyright © 2015 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.


from collections import OrderedDict

import re


NO_VALUE = type('NO_FALLBACK', (object,), {})()


class ParseError(Exception):
    pass


class UwsgiIni:
    """
    Accessor for a uwsgi ini file. The main feature of these files is that keys
    can occur more than once in sections. Furthermore, keys are not required to
    have values. Such keys will automatically receive the value `True`.

    The following is perfectly valid, for example::

        [overlord]
        daemonize = /tmp/overlord.log
        logdate
        master-fifo = /tmp/first.fifo
        master-fifo = /tmp/second.fifo

    Handling of sections remains the same though: every section must have a
    unique name.
    """

    def __init__(self):
        self.sections = OrderedDict()

    def load(self, fp):
        """
        Loads the configuration from given :term:`file object` *fp*.
        """
        return self.loads(fp.read())

    def loads(self, string):
        """
        Loads the configuration from given *string*.
        """
        self.sections = OrderedDict()
        section = None
        for i, line in enumerate(string.split('\n')):
            line = line.strip()
            if not line or line.startswith(';'):
                continue
            if line[0] == '[':
                if line[-1] != ']':
                    raise ParseError('Line %d starts with bracket, '
                                     'but doesn\'t end in one' % i)
                name = line[1:-1].strip()
                if not re.match(r'[a-zA-Z0-9_-]+$', name):
                    raise ParseError('Line %d contains invalid section name: %s'
                                     % (i, name))
                section = UwsgiSection(name)
                self.sections[name] = section
                continue
            if section is None:
                raise ParseError('Line %d is outside of any sections' % i)
            try:
                k, v = line.split('=', 1)
                k = k.strip()
                v = v.strip()
            except ValueError:
                k, v = (line, True)
            section[k] = v

    def write(self, fp):
        """
        Writes this configuration to given :term:`file object` *fp*.
        """
        fp.write(self.dumps())

    def dumps(self):
        """
        Converts this configuration to string.
        """
        result = ''
        for name, section in self.sections.items():
            result += '[%s]\n' % name
            for k, v in section:
                result += '%s = %s\n' % (k, str(v))
            result += '\n'
        return result

    def __getitem__(self, key):
        if key not in self.sections:
            self.sections[key] = UwsgiSection(key)
        return self.sections[key]

    def __setitem__(self, key, value):
        self.sections[key] = section = UwsgiSection(key)
        for k, v in value:
            section[k] = v

    def __delitem__(self, key):
        self.sections.pop(key, None)

    def __iter__(self):
        return iter(self.sections)


class UwsgiSection:
    """
    A named section within a :class:`.UwsgiIni`. Behaves very much like an
    :class:`collections.OrderedDict`, but can contain keys multiple times.
    """

    def __init__(self, name):
        self.name = name
        self.pairs = []

    def __iter__(self):
        return iter(self.pairs)

    def __setitem__(self, key, value):
        self.pairs.append((key, value))

    def __getitem__(self, key):
        for k, v in reversed(self.pairs):
            if key == k:
                return v
        raise KeyError(k)

    def __delitem__(self, key):
        self.reset(key)

    def pop(self, key, fallback=NO_VALUE):
        for i, (k, v) in reversed(enumerate(self.pairs)):
            if k == key:
                del self.pairs[i]
                return v
        if fallback != NO_VALUE:
            return fallback
        raise KeyError(key)

    def reset(self, key, value=NO_VALUE):
        """
        Removes all values with given *key*.
        """
        for i, (k, v) in reversed(enumerate(self.pairs)):
            if key == k:
                del self.pairs[i]
        if value != NO_VALUE:
            self[k] = v

    def get_all(self, key):
        """
        Provides all values for given *key*.
        """
        return [v for k, v in self.pairs if k == key]



# Copyright © 2015 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.


import click
import json
import score.uwsgi


def parse_alias(alias):
    result = alias.split('/')
    if len(result) == 1:
        result.append(None)
    return result


def zergling_status(zergling):
    status = []
    try:
        if zergling.is_reloading():
            status.append('reloading')
        if zergling.is_paused():
            status.append('paused')
    except score.uwsgi.NotRunning:
        if zergling.is_starting():
            status.append('starting')
        else:
            status.append('stopped')
    return status


@click.group()
@click.argument('conf', type=click.Path(file_okay=True, dir_okay=False))
@click.pass_context
def main(ctx, conf):
    """
    Manages uwsgi processes.
    """
    ctx.obj = score.init.init_from_file(conf)


@main.command('status')
@click.pass_context
def status(ctx):
    for overlord in ctx.obj.uwsgi.Overlord.instances():
        print(overlord.name)
        for zergling in overlord.zerglings():
            status = zergling_status(zergling)
            status = ' (%s)' % ', '.join(status) if status else ''
            print("    %s%s" % (zergling.name, status))


@main.command('spawn-overlord')
@click.argument('name')
@click.pass_context
def spawn_overlord(ctx, name):
    """
    Starts a master server that accepts zerglings.
    """
    overlord = ctx.obj.uwsgi.Overlord(name)
    overlord.regenini()
    overlord.start()


@main.command('slay-overlord')
@click.argument('name')
@click.pass_context
def slay_overlord(ctx, name):
    """
    Stops a previously started master server.
    """
    ctx.obj.uwsgi.Overlord(name).stop()


@main.command('spawn-zergling')
@click.argument('overlord')
@click.argument('file',
                type=click.Path(file_okay=True, dir_okay=False))
@click.option('-e', '--virtualenv',
              type=click.Path(file_okay=False, dir_okay=True, writable=False),
              help="Path to the virtualenv")
@click.option('-p', '--paused', is_flag=True, default=False)
@click.pass_context
def spawn_zergling(ctx, overlord, file, paused, virtualenv=None):
    overlord, name = parse_alias(overlord)
    overlord = ctx.obj.uwsgi.Overlord(overlord)
    zerglings = overlord.zerglings()
    if not name:
        maxname = len(zerglings)
        for zergling in zerglings:
            try:
                maxname = max(maxname, int(zergling.name))
            except ValueError:
                pass
        name = str(maxname + 1)
    try:
        zergling = next(z for z in zerglings if z.name == name)
    except StopIteration:
        zergling = ctx.obj.uwsgi.Zergling(overlord, name, file)
    zergling.regenini(startpaused=paused, virtualenv=virtualenv)
    zergling.start()


@main.command('pause-zergling')
@click.argument('zergling')
@click.pass_context
def pause_zergling(ctx, zergling):
    overlord, zergling = parse_alias(zergling)
    try:
        ctx.obj.uwsgi.Overlord(overlord).zergling(zergling).pause()
    except score.uwsgi.NoSuchZergling:
        raise click.ClickException('No zergling with that name.')
    except score.uwsgi.AlreadyRunning:
        raise click.ClickException('That zergling is already paused')


@main.command('stats-zergling')
@click.argument('zergling')
@click.pass_context
def stats_zergling(ctx, zergling):
    overlord, zergling = parse_alias(zergling)
    try:
        overlord = ctx.obj.uwsgi.Overlord(overlord)
        stats = overlord.zergling(zergling).read_stats()
        print(json.dumps(stats, sort_keys=True, indent=4))
    except score.uwsgi.NoSuchZergling:
        raise click.ClickException('No zergling with that name.')
    except score.uwsgi.NotRunning:
        raise click.ClickException('Zergling not running.')


@main.command('resume-zergling')
@click.argument('zergling')
@click.pass_context
def resume_zergling(ctx, zergling):
    overlord, zergling = parse_alias(zergling)
    try:
        ctx.obj.uwsgi.Overlord(overlord).zergling(zergling).resume()
    except score.uwsgi.NoSuchZergling:
        raise click.ClickException('No zergling with that name.')
    except score.uwsgi.AlreadyRunning:
        raise click.ClickException('That zergling is already running')


@main.command('kill-zergling')
@click.argument('zergling')
@click.pass_context
def kill_zergling(ctx, zergling):
    overlord, zergling = parse_alias(zergling)
    try:
        zergling = ctx.obj.uwsgi.Overlord(overlord).zergling(zergling)
        zergling.stop()
    except score.uwsgi.NoSuchZergling:
        raise click.ClickException('No zergling with that name.')
    except score.uwsgi.NotRunning:
        pass
    zergling.delete()


@main.command('reload-zergling')
@click.argument('zergling')
@click.pass_context
def reload_zergling(ctx, zergling):
    overlord, zergling = parse_alias(zergling)
    try:
        ctx.obj.uwsgi.Overlord(overlord).zergling(zergling).reload()
    except score.uwsgi.NoSuchZergling:
        raise click.ClickException('No zergling with that name.')
    except score.uwsgi.AlreadyReloading:
        raise click.ClickException('That zergling is already reloading')


if __name__ == '__main__':
    main()



# Copyright © 2015 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.


from .iniparser import UwsgiIni

import json
import logging
import os
import re
import socket
from subprocess import Popen, PIPE, DEVNULL
import sys
import textwrap

log = logging.getLogger(__name__)


class NoSuchZergling(Exception):
    """
    Thrown when the requested zergling does not exist.
    """


class AlreadyPaused(Exception):
    """
    Thrown when a paused instance is intructed to pause.
    """


class AlreadyRunning(Exception):
    """
    Thrown when a running instance is intructed to start/resume.
    """


class AlreadyReloading(Exception):
    """
    Thrown when a reloading instance is intructed to reloading.
    """


class NotRunning(Exception):
    """
    Indicates that the instance was not running.
    """


class UwsgiProcess:
    """
    A uwsgi process managed by this module.
    """

    def __init__(self):
        self._pid = None

    def start(self, *, quiet=False, checkrunning=True):
        """
        Starts this instance.

        Will print the output of the uwsgi process to stdout and stderr, unless
        *quiet* was `True`.

        If *checkrunning* is left at its default value, the function will first
        check if the instance is already running and raise
        :class:`.AlreadyRunning` if it was.
        """
        if checkrunning and self.is_running():
            raise AlreadyRunning(str(self))
        self._pid = None
        os.makedirs(self.folder, exist_ok=True)
        log.info('Starting %s' % str(self))
        stdout, stderr = sys.stdout, sys.stderr
        if quiet:
            stdout = DEVNULL
            stderr = PIPE
        process = Popen(self.cmdline, stdout=stdout, stderr=stderr,
                        cwd=os.path.join(self.folder, '..'))
        _, err = process.communicate()
        if process.returncode:
            msg = 'Error starting process %s' % self
            if err:
                msg += ':\n' + textwrap.indent(str(err, 'UTF-8'), '  ')
            raise Exception(msg)

    def stop(self):
        """
        Stops this instance.

        Will raise :class:`.NotRunning`, if the instance was already stopped.
        """
        if not self.is_running():
            raise NotRunning(str(self))
        log.info('Stopping %s' % str(self))
        open(self.fifo, 'w').write('q')
        self._pid = None

    def is_running(self):
        """
        Whether this instance is running, i.e. the process exists and is
        responding to its requests on its statistics socket. See
        :meth:`.read_stats`.
        """
        try:
            self.read_stats()
            return True
        except NotRunning:
            return False

    @property
    def pid(self):
        """
        The process id of this process, or `None` if it is not :meth:`.running`.
        """
        try:
            return self.read_stats()['pid']
        except NotRunning:
            return None

    def read_stats(self):
        """
        Reads and parses all data from this process' `statistics socket`_.

        .. _statistics socket:
            http://uwsgi-docs.readthedocs.org/en/latest/StatsServer.html
        """
        try:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect(self.stats_socket)
            result = b''
            read = sock.recv(4096)
            while read:
                result += read
                read = sock.recv(4096)
            return json.loads(str(result, 'UTF-8'))
        except (FileNotFoundError,
                ConnectionRefusedError,
                ConnectionResetError):
            raise NotRunning(str(self))


class Overlord(UwsgiProcess):
    """
    The overlord process handling client connections.
    """

    def __init__(self, name):
        super().__init__()
        self.name = name
        self.folder = os.path.join(self.conf.rootdir, self.name)
        self.fifo = os.path.join(self.folder, 'overlord.fifo')
        self.logfile = os.path.join(self.folder, 'overlord.log')
        self.stats_socket = os.path.join(self.folder, 'overlord.stats.sock')
        self.inifile = os.path.join(self.folder, 'uwsgi.ini')
        self.cmdline = ["uwsgi", "--ini", "%s/uwsgi.ini:overlord" % self.name]

    def regenini(self):
        """
        Re-generates and writes this overlord's ini file.
        """
        ini = UwsgiIni()
        section = ini['overlord']
        section['master'] = True
        section['daemonize'] = self.logfile
        section['stats-server'] = self.stats_socket
        section['plugin'] = 'zergpool'
        section['logdate'] = True
        section['zerg-pool'] = '%s:%s' % (
            os.path.join(self.folder, 'zerg.socket'),
            os.path.join(self.folder, 'socket'))
        section['master-fifo'] = self.fifo
        os.makedirs(self.folder, exist_ok=True)
        ini.write(open(self.inifile, 'w'))

    def zerglings(self):
        """
        All :class:`Zerglings <.Zergling>` associated with this overlord.
        """
        zerglings = []
        ini = UwsgiIni()
        if os.path.exists(self.inifile):
            ini.load(open(self.inifile))
        for section_name in ini:
            match = re.match(r'^zergling-(.*)$', section_name)
            if not match:
                continue
            name = match.group(1)
            section = ini[section_name]
            zergling = self.conf.Zergling._from_section(self, name, section)
            zerglings.append(zergling)
        return zerglings

    def zergling(self, name):
        try:
            return next(z for z in self.zerglings() if z.name == name)
        except StopIteration:
            raise NoSuchZergling(name)

    def __str__(self):
        return self.name


class Zergling(UwsgiProcess):

    @classmethod
    def _from_section(cls, overlord, name, section):
        return cls(overlord, name, section['ini-paste'])

    def __init__(self, overlord, name, appini):
        super().__init__()
        self.overlord = overlord
        self.name = name
        self.appini = appini
        self.folder = self.overlord.folder
        self.fifo = os.path.join(self.folder, 'zergling-%s.fifo' % name)
        self.logfile = os.path.join(self.folder, 'zergling-%s.log' % name)
        self.stats_socket = os.path.join(self.folder,
                                         'zergling-%s.stats.sock' % name)
        self.startup_file = os.path.join(self.folder,
                                         'zergling-%s.startup' % name)
        self.inifile = self.overlord.inifile
        self.cmdline = [
            "uwsgi", "--ini",
            "%s/uwsgi.ini:zergling-%s" % (self.overlord.name, self.name)]

    def regenini(self, startpaused=False, virtualenv=None):
        """
        Re-generates and updates this zerglings section in the overlord's ini
        file.  It is possible to create the configuration in a way that pauses
        the process immediately upon starting by passign a truthy value for
        *startpaused*.
        """
        ini = self._open_ini()
        if 'zergling-%s' % self.name in ini:
            del ini['zergling-%s' % self.name]
        section = ini['zergling-%s' % self.name]
        if virtualenv:
            section['virtualenv'] = virtualenv
        section['zerg'] = os.path.join(self.folder, 'zerg.socket')
        section['daemonize'] = self.logfile
        section['logdate'] = True
        section['stats-server'] = self.stats_socket
        section['master-fifo'] = self.fifo
        section['master-fifo'] = self.fifo + '.restart'
        if startpaused:
            section['plugin'] = "startpaused"
        section['plugin'] = "python%s" % ''.join(map(str, sys.version_info[:2]))
        section['ini-paste'] = self.appini
        section['hook-asap'] = 'write:%s true' % self.startup_file
        section['hook-accepting1-once'] = 'unlink:%s' % self.startup_file
        section['hook-as-user-atexit'] = 'unlink:%s.restart' % self.fifo
        section['hook-as-user-atexit'] = 'unlink:%s' % self.startup_file
        os.makedirs(self.folder, exist_ok=True)
        ini.write(open(self.inifile, 'w'))

    def reload(self, quiet=True):
        """
        Starts another instance of this process, which will stop the old
        instance once it is up and running. This means that there will be twice
        the number of processes for a short time and it will not be clear which
        of these instances will respond to an incoming request for a very brief
        time frame.
        """
        if self.is_reloading():
            raise AlreadyReloading(str(self))
        log.info('Reloading %s' % str(self))
        open(self.fifo, 'w').write('1')
        try:
            startpaused = self.is_paused()
        except NotRunning:
            startpaused = False
        ini = self._open_ini()
        section = ini['zergling-%s' % self.name]
        plugins = []
        found = False
        for plugin in section.get_all('plugin'):
            if plugin == "startpaused":
                found = True
                if startpaused:
                    break
            else:
                plugins.append(plugin)
        if found and not startpaused:
            section.reset("plugin")
            for plugin in plugins:
                section["plugin"] = plugin
        elif not found and startpaused:
            section["plugin"] = "startpaused"
        hooks = ("writefifo:%s.restart q" % self.fifo,)
        for hook in hooks:
            if hook not in section.get_all('hook-accepting1-once'):
                section['hook-accepting1-once'] = hook
        ini.write(open(self.inifile, 'w'))
        self.start(quiet=quiet, checkrunning=False)

    def pause(self):
        """
        Pauses this instance. Raises :class:`.AlreadyPaused` if already paused.
        """
        if self.is_paused():
            raise AlreadyPaused(str(self))
        log.info('Pausing %s' % str(self))
        open(self.fifo, 'w').write('p')

    def resume(self):
        """
        Resumes this instance. Raises :class:`.AlreadyRunning` if already
        running.
        """
        if not self.is_paused():
            raise AlreadyRunning(str(self))
        log.info('Resuming %s' % str(self))
        open(self.fifo, 'w').write('p')

    def start(self, *args, **kwargs):
        """
        Starts this instance. All arguments are passed to :meth:`.start`. But
        raises :class:`.AlreadyRunning` if already running.
        """
        if self.is_starting():
            raise AlreadyRunning(str(self))
        super().start(*args, **kwargs)

    def delete(self):
        """
        Removes this zergling configuration from the ini file.
        """
        if not os.path.exists(self.inifile):
            return
        ini = self._open_ini()
        if 'zergling-%s' % self.name not in ini:
            return
        del ini['zergling-%s' % self.name]
        ini.write(open(self.inifile, 'w'))
        files = (self.stats_socket, self.startup_file,
                 self.fifo, self.fifo + '.restart')
        for file in files:
            try:
                os.remove(file)
            except FileNotFoundError:
                pass

    def is_reloading(self):
        """
        Checks whether this uwsgi process is currently :meth:`reloading
        <.reload>`.
        """
        return os.path.exists(self.fifo + '.restart')

    def is_starting(self):
        """
        Whether this process is currently starting up. This value will only be
        `True` during the short time frame between the call to :meth:`.start`
        and the point at which it is :attr:`.running`.
        """
        return os.path.exists(self.startup_file)

    def is_paused(self):
        """
        Whether the instance is currently paused.
        """
        return self.read_stats()['workers'][0]['status'] == 'pause'

    def _open_ini(self):
        """
        Returns the overlords ini file as :class:`UwsgiIni` object.
        """
        ini = UwsgiIni()
        if os.path.exists(self.inifile):
            ini.load(open(self.inifile))
        return ini

    def __str__(self):
        return '%s/zergling-%s' % (self.overlord.name, self.name)
